unit PXUnit;

{$MODE Delphi}

interface

uses
  LCLIntf, LCLType, LMessages, Messages, SysUtils, Classes, Graphics, Controls, Forms, ExtCtrls,
  Dialogs, ComCtrls, StdCtrls, FileUtil, synaser;

const
  AlteraSize         = 1 shl 24;
  MemorySize         = 1 shl 24;
  OutSize            = 1 shl 24;
  TxBuffSize         = $4000;
  RxBuffSize         = $4000;

type
  TPXForm = class(TForm)
    FileLabel        : TLabel;
    SizeLabel        : TLabel;
    ProgressBar      : TProgressBar;
    procedure        FormActivate(Sender: TObject);
    procedure        LoadFile(const Param: string; var FileName: string; var Length: cardinal; Buff: PByteArray; Size: cardinal);
    procedure        TalkToPX;
    procedure        EnterFile(Buff: PByteArray; Length: cardinal);
    procedure        EnterNormal(Length, Start: cardinal; Buff: PByteArray);
    procedure        EnterRepeat(Length, Value: cardinal);
    procedure        EnterByte(Value: byte);
    procedure        ConnectToPX;
    procedure        OpenComm;
    procedure        TByte(b: byte);
    procedure        TComm;
    function         RByte: byte;
    procedure        RComm;
    procedure        CloseComm;
    function         IterateLFSR: byte;
    procedure        ErrorAbort(s: string);
  end;

var
  PXForm             : TPXForm;

  Show_Help          : boolean;
  FlashMode          : boolean;

  AlteraLength       : cardinal;
  AlteraBuff         : array [0..AlteraSize-1] of byte;

  MemoryLength       : cardinal;
  MemoryBuff         : array [0..MemorySize-1] of byte;

  OutLength          : cardinal;
  OutBuff            : array [0..OutSize-1] of byte;

  CommPort           : string;
  CommOpen           : boolean;
//  CommHandle         : THandle;
//  CommDCB            : TDCB;
  ser : TBlockSerial;
  CommTimeout        : cardinal;

  TxBuffLength       : cardinal;
  TxBuff             : array[0..TxBuffSize-1] of byte;
  RxBuffStart        : cardinal;
  RxBuffEnd          : cardinal;
  RxBuff             : array[0..RxBuffSize-1] of byte;

  LFSR               : byte;

implementation

uses PXHelp;

{$R *.lfm}

/////////////////////
//  Form Activity  //
/////////////////////

procedure TPXForm.FormActivate(Sender: TObject);
var
  i: cardinal;
  s, AlteraFileName, MemoryFileName: string;
begin
  Show_Help := True;
  FlashMode := False;
  CommPort := 'COM1';
  AlteraLength := 0;
  MemoryLength := 0;
  MemoryFileName := '';
  AlteraFileName := '';
  try
    // Check parameters for /P, /#, and valid file(s)
    if ParamCount = 0 then Abort;
    for i := 1 to ParamCount do
    begin
//      s := UpperCase(ParamStr(i));
      s := ParamStr(i);
{      if (Length(s) = 2) and (s[1] = '/') then
      begin
        if s[2] = 'P' then FlashMode := True
        else if (Byte(s[2]) >= Byte('1')) and (Byte(s[2]) <= Byte('9')) then CommPort := 'COM' + s[2]
        else ErrorAbort('Invalid /? parameter.');
      end}
      if (s[1] = '/') then
      begin
        if s[2] = 'P' then
          FlashMode := True
        else
          begin
//            writeln(s);
            CommPort := s[2..length(s)];
//            writeln(CommPort);
            if UpperCase(CommPort[1..3]) <> 'COM' then
              CommPort := '/dev/' + CommPort;
//            writeln(CommPort);
          end;
      end
      else if FileExistsUTF8(s) { *Converted from FileExists* } then
        if UpperCase(ExtractFileExt(s)) = '.RBF' then
          LoadFile(s, AlteraFileName, AlteraLength, @AlteraBuff, AlteraSize)
        else
          LoadFile(s, MemoryFileName, MemoryLength, @MemoryBuff, MemorySize)
      else ErrorAbort('File ' + s + ' not found.');
    end;
    // Make sure at least one file loaded
    if (AlteraLength = 0) and (MemoryLength = 0) then ErrorAbort('No file data to download.');
    // Make sure /P accompanied by .RBF file
    if (AlteraLength = 0) and FlashMode then ErrorAbort('An .RBF file must be specified when using /P.');
    // Update filename(s) on form
    if AlteraLength <> 0 then
      if MemoryLength <> 0 then FileLabel.Caption := AlteraFileName + ' + ' + MemoryFileName
      else FileLabel.Caption := AlteraFileName
    else FileLabel.Caption := MemoryFileName;
    Repaint;
    // Perform load/program
    ser := TBlockSerial.Create;
    ser.RaiseExcept := True;
    ConnectToPx;
    TalkToPX;
    Show_Help := False;
  finally
    // Possibly show help dialog, then close
    if Show_Help then PXHelpForm.ShowModal;
    Close;
  end;
end;

////////////////
//  File I/O  //
////////////////

// Load memory file
procedure TPXForm.LoadFile(const Param: string; var FileName: string; var Length: cardinal; Buff: PByteArray; Size: cardinal);
var
  f: file;
  s: cardinal;
begin
  FileName := ExtractFileName(Param);
  FileMode := 0;
  AssignFile(f, Param);
  try
    Reset(f, 1);
    s := FileSize(f);
    if s > Size then ErrorAbort('File ' + FileName + ' exceeds application capacity of ' + IntToStr(Size shr 20) + ' MB.');
    BlockRead(f, Buff^, s);
    Length := s;
  finally
    CloseFile(f);
  end;
end;


///////////////////
//  PX Routines  //
///////////////////

// Connect to PX
procedure TPXForm.ConnectToPX;
begin
  CommTimeout := 1000;
  OpenComm;
  // Send connect string
  TByte($FF);
  TByte($FF);
  LFSR := $FF;
  repeat TByte(IterateLFSR and 1)
  until LFSR = $FF;
  TComm;
  // Close and reopen comm to flush any errors
  CloseComm;
  OpenComm;
  // Send byte to signify ready to receive
  TByte(0);
  TComm;
  // Receive connect string
  repeat if RByte <> IterateLFSR and 1 then ErrorAbort('PX Loader not found on ' + CommPort + '.')
  until LFSR = $FF;
  // Receive version
  RByte;
end;

// Talk to PX
procedure TPXForm.TalkToPX;
var
  i: cardinal;
begin
  OutLength := 0;
  // If flash mode, enter flag byte
  if FlashMode then EnterByte($A5);
  // Enter file(s) data
  if AlteraLength <> 0 then
  begin
    EnterByte(1);
    EnterFile(@AlteraBuff, AlteraLength);
    if MemoryLength <> 0 then
    begin
      EnterByte(1);
      EnterFile(@MemoryBuff, MemoryLength);
    end
    else EnterByte(0);
  end
  else
  begin
    EnterByte(0);
    EnterFile(@MemoryBuff, MemoryLength);
  end;
  // If flash mode, extend data to 16KB boundary
  if FlashMode then while (OutLength and $3FFF) <> 0 do EnterByte(0);
  // Update data size on form
  SizeLabel.Caption := IntToStr(OutLength shr 10) + ' KB';
  Repaint;
  // Send command
  TByte(Ord(FlashMode));
  // If flash mode, extend timeout
  if FlashMode then CommTimeout := 4000;
  // Send data
  for i := 0 to OutLength - 1 do
  begin
    // If flash mode, handle 16KB program begin and 64KB sector-erase wait
    if FlashMode then
    begin
      if i and $3FFF = 0 then TByte(1);
      if i and $FFFF = 0 then
      begin
        TComm;
        RByte;
      end;
    end;
    // Send data byte
    TByte(OutBuff[i]);
    // if flash mode, handle 16KB program wait
    if FlashMode then
      if i and $3FFF = $3FFF then
      begin
        TComm;
        RByte;
      end;
    // Update progress indicator
    if i mod (OutLength shr 6 + 1) = 0 then
    begin
      ProgressBar.Position := i shl 5 div OutLength;
      ProgressBar.Repaint;
    end;
  end;
  // Terminate PX connection according to mode
  if FlashMode then
  begin
    TByte(0);
    TComm;
  end
  else
  begin
    TComm;
    i := RByte;
    TByte(0);
    TComm;
    if i <> 1 then ErrorAbort('Altera device is not configured properly.');
  end;
  CloseComm;
end;


/////////////////////
//  Data Compiler  //
/////////////////////

// Enter file as packets
procedure TPXForm.EnterFile(Buff: PByteArray; Length: cardinal);
var
  NormalCount, RepeatCount, RepeatValue: byte;
  NormalMode: boolean;
  i: cardinal;
begin
  // Init packet compressor
  NormalCount := 0;
  RepeatCount := 0;
  RepeatValue := Buff[0] xor $FF;
  NormalMode := True;
  // Enter packets
  for i := 0 to Length-1 do
  begin
    if NormalMode then
    begin
      if Buff[i] <> RepeatValue then
      begin
        Inc(NormalCount);
        RepeatCount := 1;
        RepeatValue := Buff[i];
        if NormalCount >= 120 then
        begin
          EnterNormal(NormalCount, i + 1 - NormalCount, Buff);
          NormalCount := 0;
          RepeatCount := 0;
        end;
      end
      else
      begin
        Inc(NormalCount);
        Inc(RepeatCount);
        if RepeatCount = 3 then
        begin
          EnterNormal(NormalCount - RepeatCount, i + 1 - NormalCount, Buff);
          NormalMode := False;
        end;
      end;
    end
    else
    begin
      if Buff[i] = RepeatValue then
      begin
        Inc(RepeatCount);
        if RepeatCount = 120 then
        begin
          EnterRepeat(RepeatCount, RepeatValue);
          NormalCount := 0;
          RepeatCount := 0;
          NormalMode := True;
        end;
      end
      else
      begin
        EnterRepeat(RepeatCount, RepeatValue);
        NormalCount := 1;
        RepeatCount := 1;
        RepeatValue := Buff[i];
        NormalMode := True;
      end;
    end;
  end;
  // Enter any final packet
  if NormalMode then
    EnterNormal(NormalCount, Length - NormalCount, Buff)
  else
    EnterRepeat(RepeatCount, RepeatValue);
  // Enter end byte
  EnterByte(0);
end;

// Enter normal data packet
procedure TPXForm.EnterNormal(Length, Start: cardinal; Buff: PByteArray);
var
  i: cardinal;
begin
  if Length = 0 then Exit;
  EnterByte(Length);
  for i := 0 to Length - 1 do EnterByte(Buff[Start + i]);
end;

// Enter repeat data packet
procedure TPXForm.EnterRepeat(Length, Value: cardinal);
begin
  EnterByte($100 - Length);
  EnterByte(Value);
end;

// Enter byte
procedure TPXForm.EnterByte(Value: byte);
begin
  if OutLength = OutSize then
  begin
    TByte(0);
    TByte(0);
    TByte(0);
    TComm;
    RByte;
    TByte(0);
    TComm;
    ErrorAbort('Data exceeds application capacity of ' + IntToStr(OutSize shr 20) + ' MB.')
  end;
  OutBuff[OutLength] := Value;
  Inc(OutLength);
end;


/////////////////////
//  Comm Routines  //
/////////////////////

// Open comm port
procedure TPXForm.OpenComm;
{const
  CommTimeouts: TCOMMTIMEOUTS = (
    ReadIntervalTimeout: MAXDWORD;
    ReadTotalTimeoutMultiplier: 0;
    ReadTotalTimeoutConstant: 0;
    WriteTotalTimeoutMultiplier: 0;
    WriteTotalTimeoutConstant: 0);}
var
  Ticks: cardinal;
begin
  Ticks := GetTickCount;
  while GetTickCount - Ticks < 50 do
    Sleep(0);
  CommOpen := False;
{  CommHandle := CreateFile(PChar('\\.\' + CommPort), GENERIC_READ or GENERIC_WRITE,
    0, nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, cardinal(0));
  if CommHandle = INVALID_HANDLE_VALUE then
    ErrorAbort('Unable to open ' + CommPort + '.');
  BuildCommDCB(PChar('2000000,n,8,1'), CommDCB);
  CommDCB.Flags := $0;
  SetCommState(CommHandle, CommDCB);
  SetCommTimeouts(CommHandle, CommTimeouts);}
  ser.Connect(CommPort);
  // Error checking missing!!!
  ser.Config(2000000,8,'N',SB1,False,False);
  ser.SetCommState;

  TxBuffLength := 0;
  RxBuffStart := 0;
  RxBuffEnd := 0;
  CommOpen := True;
end;

// Add byte to comm buffer
procedure TPXForm.TByte(b: byte);
begin
  TxBuff[TxBuffLength] := b;
  Inc(TxBuffLength);
  if TxBuffLength = TxBuffSize then TComm;
end;

// Transmit comm buffer
procedure TPXForm.TComm;
var
  BytesWritten: cardinal;
begin
//  if not WriteFile(CommHandle, TxBuff, TxBuffLength, BytesWritten, nil) then
//    ErrorAbort('Unable to write to ' + CommPort + '.');
  BytesWritten := ser.SendBuffer(@TxBuff, TxBuffLength);
  if BytesWritten <> TxBuffLength then
    ErrorAbort('Transmit stall on ' + CommPort + '.');
  TxBuffLength := 0;
end;

// Receive comm byte
function TPXForm.RByte: byte;
begin
  if RxBuffStart = RxBuffEnd then RComm;
  Result := RxBuff[RxBuffStart];
  Inc(RxBuffStart);
end;

// Receive comm data
procedure TPXForm.RComm;
var
  Ticks: cardinal;
begin
  RxBuffStart := 0;
  Ticks := GetTickCount;
  while GetTickCount - Ticks < CommTimeout do
  begin
//    if not ReadFile(CommHandle, RxBuff, RxBuffSize, RxBuffEnd, nil) then
//      ErrorAbort('Unable to read from ' + CommPort + '.');
    RxBuffEnd := ser.RecvBuffer(@RxBuff, RxBuffSize);
    if RxBuffEnd > 0 then Exit;
  end;
  ErrorAbort('PX Loader not found on ' + CommPort + '.');
end;

// Close comm port
procedure TPXForm.CloseComm;
begin
//  FileClose(CommHandle); { *Converted from CloseHandle* }
  ser.CloseSocket;
  CommOpen := False;
end;

/////////////////////
//  Miscellaneous  //
/////////////////////

// Iterate LFSR
function TPXForm.IterateLFSR: byte;
begin
  LFSR := LFSR shl 1 and $FE or (LFSR shr 7 xor LFSR shr 3 xor LFSR shr 2 xor LFSR shr 1) and 1;
  Result := LFSR;
end;

// Error/abort
procedure TPXForm.ErrorAbort(s: string);
begin
  if CommOpen then CloseComm;
  MessageDlg(s, mtError, [mbOK], cardinal(0));
  Show_Help := False;
  Abort;
end;

end.

